from distconfig.backends.base import BaseBackend
from distconfig.backends.execution_context import ThreadingExecutionContext


class ConsulBackend(BaseBackend):
    """Consul backend implementation.

    :param client: Instance of :class:`consul.Consul`.
    :param execution_context: Instance of :class:`distconfig.backends.execution_context.ExecutionContext`

    """

    def __init__(self, client, execution_context=ThreadingExecutionContext(), **kwargs):
        super(ConsulBackend, self).__init__(**kwargs)
        self._client = client
        self._execution_context = execution_context
        self._watching = set()

    def get_raw(self, key):
        result = self._get_backend_data(key)
        self._add_watcher(key)
        return result

    def _get_backend_data(self, key):
        _, result = self._client.kv.get(key)
        if result:
            result = result['Value']
        return result

    def _add_watcher(self, key):
        if key not in self._watching:
            self._watching.add(key)
            self._execution_context.run(self._watch_for_changes, key)

    def _watch_for_changes(self, key):
        index = None
        while 1:
            try:
                index, data = self._client.kv.get(key, index=index)
            except Exception as ex:
                self._logger.error('exception raised while listening on consul changes (re-launching watcher): %s', ex)
            else:
                if data:
                    data = data['Value']
                self._notify_listeners(data)





def resolve_dotted_name(name):
    """Return object referenced by dotted name.

    :param name: dotted name as a String.
    :return: Resolved Python object.
    :raises ImportError: If can't resolve ``none``

    Examples:

        >>> resolve_dotted_name('sys.exit')
        <built-in function exit>
        >>> resolve_dotted_name('xml.etree.ElementTree')  # doctest: +ELLIPSIS
        <module 'xml.etree.ElementTree' ...>
        >>> resolve_dotted_name('distconfig.backends.zookeeper.ZooKeeperBackend')
        <class 'distconfig.backends.zookeeper.ZooKeeperBackend'>

    """
    paths = name.split('.')
    current = paths[0]
    found = __import__(current)
    for part in paths[1:]:
        current += '.' + part
        try:
            found = getattr(found, part)
        except AttributeError:
            found = __import__(current, fromlist=part)
    return found



import abc
import json
import os
import time
import unittest

envvar = 'DISTCONFIG_RUN_INTEGRATION_TEST'
if os.environ.get(envvar) != 'true':
    raise unittest.SkipTest('Skipping integration tests, to enable run: export %s=true' % envvar)


class _BackendTestCase(unittest.TestCase, metaclass=abc.ABCMeta):

    def setUp(self):
        super(_BackendTestCase, self).setUp()
        self.initial_config = {
            'string': 'foobar',
            'integer': 1,
            'float': 1.5,
            'boolean': False,
            'none': None
        }

    @property
    @abc.abstractmethod
    def path(self):
        pass

    @property
    @abc.abstractmethod
    def proxy(self):
        pass

    @abc.abstractmethod
    def create_config(self, value):
        pass

    @abc.abstractmethod
    def update_config(self, new_value):
        pass

    @abc.abstractmethod
    def delete_config(self):
        pass

    def serialize_config(self, config):
        return json.dumps(config).encode('utf8')

    def wait_for_backend_changes(self, timeout=5):
        config_value = None

        def callback(new_value):
            config_value = new_value  # noqa

        self.proxy.backend.add_listener(callback)

        start = time.time()
        while config_value is None and (time.time() - start) < timeout:
            time.sleep(1)

    def test_get_config(self):
        self.create_config(self.initial_config)

        config = self.proxy.get_config(self.path)

        self.assertDictEqual(config._data, self.initial_config)

    def test_get_unexistant_config(self):
        config = self.proxy.get_config('unexistant-path')

        self.assertDictEqual(config._data, {})

    def test_get_changed_config(self):
        self.create_config(self.initial_config)

        config = self.proxy.get_config(self.path)

        new_config = {
            'new': 'config',
            'different': 'values',
        }
        self.update_config(new_config)

        self.wait_for_backend_changes()

        self.assertDictEqual(config._data, new_config)

    def test_get_deleted_config(self):
        self.create_config(self.initial_config)

        config = self.proxy.get_config(self.path)

        self.delete_config()

        self.wait_for_backend_changes()

        self.assertDictEqual(config._data, {})

    def test_get_changed_two_times_config(self):
        self.create_config(self.initial_config)

        config = self.proxy.get_config(self.path)

        self.update_config({
            'firstname': 'Mouad',
            'lastname': 'Benchchaoui',
        })

        self.update_config(self.initial_config)

        self.wait_for_backend_changes()

        self.assertDictEqual(config._data, self.initial_config)

    def test_delete_and_create_config(self):
        config = self.proxy.get_config(self.path)

        self.create_config(self.initial_config)

        self.wait_for_backend_changes()

        self.assertDictEqual(config._data, self.initial_config)
